# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import logging
import os
import numpy as np
import shutil
from scipy.signal import convolve2d
from osgeo import gdal
from equi7grid.equi7grid import Equi7Grid
from equi7grid.image2equi7grid import image2equi7grid

# biomassL2 processor imports
from biopal.fh.processing_FH import (
    estimate_height,
    heigths_masking_and_merging,
)
from biopal.data_operations.data_operations import (
    read_and_oversample_data,
    read_and_oversample_aux_data,
    fnf_equi7_load_filter_equi7format,
    fnf_tandemx_load_filter_equi7format,
    fnf_and_validity_masks_merging,
    apply_dem_flattening,
    mosaiking,
)
from biopal.utility.utility_functions import (
    Task,
    set_gdal_paths,
    choose_equi7_sampling,
    check_if_path_exists,
    check_if_geometry_auxiliaries_are_present,
    check_fnf_folder_format,
    get_min_time_stamp_repository,
    resolution_heading_correction,
    decode_unique_acquisition_id_string,
    evaluate_estimation_quality_matrix,
    check_equi7_mask_coverage,
    save_breakpoints,
    collect_stacks_to_be_merged,
)
from biopal.geocoding.geocoding import (
    geocoding,
    geocoding_init,
)
from biopal.io.xml_io import (
    parse_input_file,
    parse_configuration_file,
    write_input_file,
    core_processing_fh,
)
from biopal.io.data_io import tiff_formatter
from biopal.screen_calibration.screen_calibration import apply_calibration_screens
from biopal.geometry.utility_geometry import compute_and_oversample_geometry_auxiliaries


class ForestHeight(Task):
    """FH main APP ForestHeight"
    
    run this APP to execute the complete Forest Height estimation processing chain.

    ForestHeight is composed by two sub APPS automatically called in sequence  
    when standard launch is performed:
    StackBasedProcessingFH -> CoreProcessingFH
   
    Refer to dataset_query, StackBasedProcessingFH and CoreProcessingFH documentation for step by step run.

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file
        
    Methods
    -------
    run( input_file_path )
        run the ForestHeight processing
    name : str
        name of the APP

    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query: it's the APP to be called before this APP
    StackBasedProcessingFH : it's the first of the two sub-APPs called by ForestHeight preocessor
    CoreProcessingFH : it's  the second of the two sub-APPs called by ForestHeight processor
    
    Examples
    --------
    Manual FH chain execution

    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.fh.main_FH import ForestHeight
    >>> dq_obj = dataset_query()
    >>> input_file_up = dq_obj.run( input_file )
    >>> chain_obj = ForestHeight( configuration_file )
    >>> chain_obj.run( input_file_up )

    - input_file: path of the BioPAL input file
    - input_file_up: same of input_file with also the "stack_based_processing" section
    - configuration_file: path of the BioPAL configuration file
    """

    def __init__(self, configuration_file):
        super().__init__(configuration_file)

    def _run(self, input_file):

        # Main APP #1: Stack Based Processing
        stack_based_processing_obj = StackBasedProcessingFH(self.configuration_file)

        # Run Main APP #1: Stack Based Processing
        input_file_updated = stack_based_processing_obj.run(input_file)

        # Main APP #2: Core Processing
        fh_processing_obj = CoreProcessingFH(self.configuration_file)

        # Run Main APP #2: AGB Core Processing
        fh_processing_obj.run(input_file_updated)


class StackBasedProcessingFH(Task):
    """StackBasedProcessingFH APP 
    
    StackBasedProcessingFH APP is the first of the two sub-APPs called by ForestHeight processor.
    
    It performs the stack-based forest height estimation.

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file
        
    Methods
    -------
    run( input_file_path )
        run the StackBasedProcessingFH APP
    name : str
        name of the APP        

    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query : it's the first APP to be called in the manual sequence
    CoreProcessingFH : it's the core APP that follows this APP in the call sequence

    Examples
    --------
    Manual FH chain execution
        
    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.fh.main_FH import StackBasedProcessingFH
    >>> dq_obj = dataset_query()
    >>> input_file_up1 = dq_obj.run( input_file )
    >>> sbp_obj = StackBasedProcessingFH( config_file )
    >>> input_file_up2 = sbp_obj.run( input_file_up1 )
    >>> fhcp_obj = CoreProcessingFH( config_file )
    >>> fhcp_obj.run( input_file_up2 )

    - input_file: path of the BioPAL input file
    - input_file_up1: same of input_file with also the "stack_based_processing" section   
    - input_file_up2: same of input_file_up1 with also the "core_processing_fh" section
    - config_file: path of the BioPAL configuration file
    """

    def __init__(self, configuration_file):
        super().__init__(configuration_file)

    def _run(self, input_file):

        ########################## INITIAL STEPS ##############################

        logging.info("FH: Reading chains configuration files")
        check_if_path_exists(self.configuration_file, "FILE")
        conf_params_obj = parse_configuration_file(self.configuration_file)
        input_params_obj = parse_input_file(input_file)

        ### managing output folders:
        products_folder = os.path.join(input_params_obj.output_specification.output_folder, "Products")
        if conf_params_obj.processing_flags.save_breakpoints:
            breakpoints_output_folder = os.path.join(products_folder, "breakpoints")
            logging.info("FH: Breakpoints will be saved into: " + breakpoints_output_folder)
            os.makedirs(breakpoints_output_folder)

        temp_output_folder = os.path.join(products_folder, "temp")
        logging.info("FH: Temporary data folder:" + temp_output_folder + "\n")
        os.makedirs(temp_output_folder)

        ### get temporal date time of the input data (get the minimum date from all the stacks)
        time_tag_mjd_initial = get_min_time_stamp_repository(
            input_params_obj.dataset_query.L1C_repository, input_params_obj.stack_based_processing.stack_composition
        )

        ### initialize the equi7 sampling grid
        equi7_sampling = choose_equi7_sampling(
            conf_params_obj.estimate_fh.product_resolution,
            input_params_obj.output_specification.geographic_grid_sampling,
        )
        e7g = Equi7Grid(equi7_sampling)
        logging.info("EQUI7 Grid sampling used: {}".format(equi7_sampling))

        # get needed parameters from input and configuration files
        geographic_boundaries = input_params_obj.stack_based_processing.geographic_boundaries

        gdal_path, _ = set_gdal_paths(conf_params_obj.gdal.gdal_path, conf_params_obj.gdal.gdal_environment_path)

        ########################## INITIAL STEPS END ##############################

        kz_mask = {}
        data_equi7_fnames = {}
        mask_equi7_fnames = {}
        ########################## STACK BASED STEPS ##############################
        for stack_idx, (unique_stack_id, acquisitions_pf_names) in enumerate(
            input_params_obj.stack_based_processing.stack_composition.items()
        ):

            # make temporary sub-directories
            temp_output_folder_gr = os.path.join(temp_output_folder, "geocoded", unique_stack_id)
            temp_output_folder_e7 = os.path.join(temp_output_folder, "equi7", unique_stack_id)
            os.makedirs(temp_output_folder_gr)
            os.makedirs(temp_output_folder_e7)

            ### load data ( and oversample if requested and if needed )
            try:
                logging.info("FH: Data loading for stack " + unique_stack_id + "; this may take a while:")

                (beta0_calibrated, master_id, raster_info, raster_info_orig,) = read_and_oversample_data(
                    input_params_obj.dataset_query.L1C_repository,
                    acquisitions_pf_names,
                    conf_params_obj.processing_flags.enable_resampling,
                )

            except Exception as e:
                logging.error("FH: error during input data reading: " + str(e), exc_info=True)
                raise

            ### load or compute auxiliary data
            try:

                read_dist = conf_params_obj.estimate_fh.spectral_shift_filtering
                read_ref_h = (
                    not conf_params_obj.processing_flags.apply_calibration_screen
                    and conf_params_obj.processing_flags.DEM_flattening
                )
                read_cal_screens = conf_params_obj.processing_flags.apply_calibration_screen
                geometry_aux_are_present = check_if_geometry_auxiliaries_are_present(
                    input_params_obj.stack_based_processing,
                    unique_stack_id,
                    acquisitions_pf_names,
                    read_ref_h=read_ref_h,
                    read_dist=read_dist,
                )

                if conf_params_obj.processing_flags.compute_geometry or not geometry_aux_are_present:

                    # messages for the log:
                    if conf_params_obj.processing_flags.compute_geometry:
                        logging.info("FH: calling geometry library for stack " + unique_stack_id + "\n")
                        if geometry_aux_are_present:
                            logging.warning("    geometry auxiliaries will be overwritten for stack " + unique_stack_id)
                        else:
                            logging.info("\n")
                    else:
                        logging.warning(
                            'FH: calling geometry library since AuxiliaryProductsFolder "Geometry" is empty or not complete \n'
                        )

                    (
                        _,
                        _,
                        ellipsoid_slope,
                        _,
                        _,
                        _,
                        sar_geometry_master,
                    ) = compute_and_oversample_geometry_auxiliaries(
                        input_params_obj.dataset_query.L1C_repository,
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        master_id,
                        conf_params_obj.processing_flags.enable_resampling,
                        comp_ref_h=read_ref_h,
                        comp_dist=read_dist,
                        force_ellipsoid=True,
                    )

                    (
                        ecef_grid,
                        off_nadir_angle_rad,
                        slope,
                        kz,
                        reference_height,
                        R,
                        _,
                    ) = compute_and_oversample_geometry_auxiliaries(
                        input_params_obj.dataset_query.L1C_repository,
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        master_id,
                        conf_params_obj.processing_flags.enable_resampling,
                        comp_ref_h=read_ref_h,
                        sar_geometry_master=sar_geometry_master,
                    )

                    logging.info("Geometry library: correcting geometry global / local reference")
                    slope = slope - ellipsoid_slope
                    for swath_id in kz.keys():
                        kz[swath_id] = (
                            kz[swath_id]
                            * np.sin(off_nadir_angle_rad[master_id])
                            / np.sin(off_nadir_angle_rad[master_id] - ellipsoid_slope)
                        )
                    off_nadir_angle_rad[master_id] = off_nadir_angle_rad[master_id] - ellipsoid_slope
                    del ellipsoid_slope

                    logging.info("FH: ...geometry auxiliaries computation done.")

                else:

                    logging.info(
                        "FH: geometry auxiliaries are provided from user, so they are now being loaded and not computed, for stack "
                        + unique_stack_id
                        + "\n"
                    )

                    (
                        ecef_grid,
                        off_nadir_angle_rad,
                        slope,
                        kz,
                        reference_height,
                        R,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                    ) = read_and_oversample_aux_data(
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        conf_params_obj.processing_flags.enable_resampling,
                        raster_info_orig,
                        read_ref_h=read_ref_h,
                        read_dist=read_dist,
                    )
                    logging.info("FH: ...geometry auxiliaries loading done.")

                # read the rest of auxiliaries which are notpart of the geometry library:
                if read_cal_screens:

                    logging.warning("FH: loading calibration screens \n")

                    (
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        cal_screens,
                        cal_screens_raster_info,
                        _,
                        _,
                        _,
                    ) = read_and_oversample_aux_data(
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        conf_params_obj.processing_flags.enable_resampling,
                        raster_info_orig,
                        read_cal_screens=read_cal_screens,
                        read_ecef=False,
                        read_off_nadir=False,
                        read_slope=False,
                        read_kz=False,
                        read_ref_h=False,
                        read_dist=False,
                    )
                    logging.info("...done")

            except Exception as e:
                logging.error(
                    "FH: error during auxiliary data computation and/or loading: " + str(e), exc_info=True,
                )
                raise

            ### Screen calibration (ground steering)
            try:
                if conf_params_obj.processing_flags.apply_calibration_screen:
                    logging.info("FH: applying calibration screen...")
                    beta0_calibrated = apply_calibration_screens(
                        beta0_calibrated, raster_info, cal_screens, cal_screens_raster_info, master_id,
                    )
                    logging.info("...done.\n")

                elif conf_params_obj.processing_flags.DEM_flattening:
                    logging.info("FH: DEM flattening... ")
                    beta0_calibrated = apply_dem_flattening(
                        beta0_calibrated, kz, reference_height, master_id, raster_info
                    )
                    logging.info("...done.\n")

            except Exception as e:
                logging.error("FH: error during screen calibration or DEM flattening." + str(e), exc_info=True)
                raise

            ### compute mean incidence angle
            look_angle_rad = np.nanmean(off_nadir_angle_rad[master_id])  # 0.4886921905584123
            logging.info("FH: incidence angle used is {} [deg] \n".format(np.rad2deg(look_angle_rad)))

            ### mean of off nadir and slope over final resolution
            windtm_x = np.int(
                np.round(conf_params_obj.estimate_fh.product_resolution / raster_info.pixel_spacing_az / 2) * 2 + 1
            )
            windtm_y = np.int(
                np.round(
                    conf_params_obj.estimate_fh.product_resolution
                    / (raster_info.pixel_spacing_slant_rg / np.sin(look_angle_rad))
                    / 2
                )
                * 2
                + 1
            )

            logging.info(" FH: computing mean of off nadir based on final product resolution...")

            off_nadir_angle_master_filtered_rad = convolve2d(
                off_nadir_angle_rad[master_id], np.ones((windtm_y, windtm_x)) / windtm_y / windtm_x, mode="same",
            )

            logging.info("...done.")

            logging.info(" FH: computing mean of slope based on final product resolution...")

            slope_filtered = convolve2d(slope, np.ones((windtm_y, windtm_x)) / windtm_y / windtm_x, mode="same")

            logging.info("...done.")

            logging.info(" FH: computing kz mask with configuration thresholds ")
            kz_list_temp = []
            kz_nan_mask = np.zeros(kz[next(iter(kz))].shape, dtype=bool)
            if master_id not in kz.keys():
                master_id = list(kz.keys())[0]
            for acq_id in kz.keys():
                kz[acq_id] = (
                    kz[acq_id]
                    * np.sin(off_nadir_angle_master_filtered_rad)
                    / np.sin(off_nadir_angle_master_filtered_rad - slope_filtered)
                )
                kz_list_temp.append(kz[acq_id])
                kz_nan_mask = np.logical_or(kz_nan_mask, np.isnan(kz[acq_id]))

            del off_nadir_angle_master_filtered_rad
            ### creating mask for KZ:
            # this will be integrated after geocoding with
            # a second mask keeping into account estimation valid values and
            # a third map which is the forest non forest mask
            delta_kz = np.maximum.reduce(kz_list_temp) - np.minimum.reduce(kz_list_temp)
            condition_curr = np.logical_and(
                delta_kz > conf_params_obj.estimate_fh.kz_thresholds[0],
                delta_kz < conf_params_obj.estimate_fh.kz_thresholds[1],
            )
            kz_mask = np.where(condition_curr, True, False)
            kz_mask[kz_nan_mask] = False

            del kz_list_temp, kz_nan_mask

            # covariance estimation window size, it may be modified by an internal flag in case of air-plane geometry
            cov_est_window_size = conf_params_obj.estimate_fh.product_resolution

            if conf_params_obj.processing_flags.multilook_heading_correction:
                _, heading_deg, _, _, _, _ = decode_unique_acquisition_id_string(unique_stack_id + "_BSL_00")

                cov_est_window_size = resolution_heading_correction(cov_est_window_size, heading_deg)

            logging.info("Covariance Estimation Window Size used: {} [m]".format(cov_est_window_size))

            ### height estimation
            logging.info("FH: " + unique_stack_id + ": performing heigth estimation...")
            try:

                if conf_params_obj.estimate_fh.spectral_shift_filtering:

                    (
                        estimated_height,
                        extinctionmap,
                        ratiomap,
                        gammaT1map,
                        gammaT2map,
                        gammaT3map,
                        kz,
                        rg_vec_subs,
                        az_vec_subs,
                        subs_F_r,
                        subs_F_a,
                        MBMP_correlation,
                    ) = estimate_height(
                        beta0_calibrated,
                        cov_est_window_size,
                        raster_info.pixel_spacing_slant_rg,
                        raster_info.pixel_spacing_az,
                        look_angle_rad,
                        raster_info.carrier_frequency_hz,
                        raster_info.range_bandwidth_hz,
                        kz,
                        conf_params_obj.estimate_fh,
                        R,
                        off_nadir_angle_rad,
                        slope,
                    )
                else:
                    (
                        estimated_height,
                        extinctionmap,
                        ratiomap,
                        gammaT1map,
                        gammaT2map,
                        gammaT3map,
                        kz,
                        rg_vec_subs,
                        az_vec_subs,
                        subs_F_r,
                        subs_F_a,
                        MBMP_correlation,
                    ) = estimate_height(
                        beta0_calibrated,
                        cov_est_window_size,
                        raster_info.pixel_spacing_slant_rg,
                        raster_info.pixel_spacing_az,
                        look_angle_rad,
                        raster_info.carrier_frequency_hz,
                        raster_info.range_bandwidth_hz,
                        kz,
                        conf_params_obj.estimate_fh,
                    )

                estimated_height = estimated_height / np.cos(slope_filtered[rg_vec_subs, :][:, az_vec_subs])
                logging.info("...done.\n")

            except Exception as e:
                logging.error("FH: error during height estimation: " + str(e), exc_info=True)
                raise

            del beta0_calibrated, slope, slope_filtered, off_nadir_angle_rad

            ### Placemark for the quality estimation to be defined
            try:
                quality_layer_sr = evaluate_estimation_quality_matrix(estimated_height.shape)

            except Exception as e:
                logging.error("FH: error during estimation quality layer evaluation: " + str(e), exc_info=True)
                raise

            ### Interpolate it over a regular lat lon grid (with grid data): generate a regular grid for the interpolation by using Max and Min lat lon from the ECEFGRID_LLH (make the vectors a bit longer), for the grid steps use the minimum steps from the ECEFGRID_LLH)
            logging.info(unique_stack_id + ": Geocoding data...")
            try:

                # initialize the geocoding
                min_spacing_m = min(
                    subs_F_a * raster_info.pixel_spacing_az, subs_F_r * raster_info.pixel_spacing_slant_rg,
                )
                min_spacing_m = min(min_spacing_m, equi7_sampling)

                logging.info("Geocoding spacing set to {} [m]".format(min_spacing_m))

                (
                    lon_in,
                    lat_in,
                    alt_in,
                    lonMeshed_out,
                    latMeshed_out,
                    valid_values_mask,
                    lon_regular_vector,
                    used_lon_step,
                    lat_regular_vector,
                    used_lat_step,
                ) = geocoding_init(ecef_grid, rg_vec_subs, az_vec_subs, min_spacing_m)

                # geocode the FH
                logging.info(unique_stack_id + ": geocoding the estimated height...")
                data_ground = geocoding(
                    estimated_height, lon_in, lat_in, lonMeshed_out, latMeshed_out, valid_values_mask,
                )
                logging.info("...done.\n")

                # geocode the KZ mask
                logging.info(unique_stack_id + ": Geocoding kz mask...")
                kz_mask_subs = kz_mask[rg_vec_subs, :][:, az_vec_subs]
                kz_mask_ground = geocoding(
                    kz_mask_subs, lon_in, lat_in, lonMeshed_out, latMeshed_out, valid_values_mask
                )
                # round the mask to quantize it because it is boolean: NOTE, do not cast to bool  otherwise the NaN will become True!
                kz_mask_ground = np.round(kz_mask_ground)

                # geocode the FH quality layer
                logging.info(unique_stack_id + ": geocoding the estimated height quality layer...")
                quality_layer_ground = geocoding(
                    quality_layer_sr, lon_in, lat_in, lonMeshed_out, latMeshed_out, valid_values_mask,
                )
                logging.info("...done.\n")

                del kz_mask, kz_mask_subs
                logging.info("...done.\n")

            except Exception as e:
                logging.error("FH: error during geocoding: " + str(e), exc_info=True)
                raise

            ### saving breakpoints
            if conf_params_obj.processing_flags.save_breakpoints:
                logging.info("FH: saving breakpoints (in slant range geometry) on " + breakpoints_output_folder)
                post_string = "_SR_" + unique_stack_id

                breakpoint_names = [
                    "estimated_height" + post_string,
                    "extinctionmap" + post_string,
                    "ratiomap" + post_string,
                    "gammaT1" + post_string,
                    "gammaT2" + post_string,
                    "gammaT3" + post_string,
                ]

                save_breakpoints(
                    breakpoints_output_folder,
                    breakpoint_names,
                    [estimated_height, extinctionmap, ratiomap, gammaT1map, gammaT2map, gammaT3map],
                )
                logging.info("...done.\n")
            del estimated_height, kz

            ### creating mask to exclude estimation not valid values:
            condition_curr = np.logical_and(
                data_ground > conf_params_obj.estimate_fh.model_parameters.estimation_valid_values_limits[0],
                data_ground < conf_params_obj.estimate_fh.model_parameters.estimation_valid_values_limits[1],
            )
            estimation_mask_ground = np.where(condition_curr, True, False)
            estimation_mask_ground[np.isnan(data_ground)] = False

            ### logical AND with kz and estimation masks: before merging also the FNF mask will be added
            # also casted to float, for incapsulation in a geotiff
            kz_and_valid_values_mask_ground = np.logical_and(kz_mask_ground, estimation_mask_ground)
            kz_and_valid_values_mask_ground.astype(float)
            kz_and_valid_values_mask_ground = np.round(kz_and_valid_values_mask_ground)

            del kz_mask_ground, estimation_mask_ground

            ### create GEOTIFF of all the layers (estimation, mask ):
            logging.info(unique_stack_id + ": formatting data to GEOTIFF...")
            try:

                data_ground_fname = os.path.join(temp_output_folder_gr, "FH.tif")
                kz_and_valid_values_mask_ground_fname = os.path.join(temp_output_folder_gr, "fnf.tif")
                #           fnf_mask_ground_fname                 = os.path.join( temp_output_folder, 'fnf_mask_ground_'                 +unique_stack_id)

                upper_left_easting_coord = lon_regular_vector[0]  # i.e. horizontal
                sampling_step_east_west = used_lon_step
                upper_left_northing_coord = lat_regular_vector[0]  # i.e. vertical
                sampling_step_north_south = used_lat_step
                geotransform = [
                    upper_left_easting_coord,
                    sampling_step_east_west,
                    0,
                    upper_left_northing_coord,
                    0,
                    sampling_step_north_south,
                ]

                # forest height geotiff formatting
                tiff_formatter(
                    [data_ground, quality_layer_ground],
                    data_ground_fname,
                    geotransform,
                    gdal_data_format=gdal.GDT_Float32,
                    multi_layers_tiff=True,
                )

                # forest height temporary mask geotiff formatting (this mask should be merged with the fnf mask, generated after )
                tiff_formatter(
                    kz_and_valid_values_mask_ground,
                    kz_and_valid_values_mask_ground_fname,
                    geotransform,
                    gdal_data_format=gdal.GDT_Float32,
                )

                logging.info("...done.\n")

            except Exception as e:
                logging.error("FH: error during GEOTIFF formatting: " + str(e), exc_info=True)
                raise

            ### formatting data to EQUI7
            logging.info(unique_stack_id + ": formatting into EQUI7 grid...")
            try:

                equi7_sampling = choose_equi7_sampling(
                    conf_params_obj.estimate_fh.product_resolution,
                    input_params_obj.output_specification.geographic_grid_sampling,
                )
                e7g = Equi7Grid(equi7_sampling)
                logging.info("    EQUI7 Grid sampling used: {}".format(equi7_sampling))

                equi7_data_outdir = os.path.join(temp_output_folder_e7, "data",)
                equi7_temp_mask_outdir = os.path.join(temp_output_folder_e7, "fnf")

                # in general from here the Equi7 can output multiple tiles, which file names are stored in the output list ( wrapped here in a dict for the stack )
                data_equi7_fnames[unique_stack_id] = image2equi7grid(
                    e7g,
                    data_ground_fname,
                    equi7_data_outdir,
                    gdal_path=gdal_path,
                    inband=None,
                    subgrid_ids=None,
                    accurate_boundary=False,
                    withtilenamesuffix=False,
                    resampling_type="bilinear",
                    tile_nodata=np.nan,
                )
                validity_mask_equi7_fnames = image2equi7grid(
                    e7g,
                    kz_and_valid_values_mask_ground_fname,
                    equi7_temp_mask_outdir,
                    gdal_path=gdal_path,
                    inband=None,
                    subgrid_ids=None,
                    accurate_boundary=False,
                    withtilenamesuffix=False,
                    resampling_type="bilinear",
                    tile_nodata=np.float(0),
                )

            except Exception as e:
                logging.error("FH: error during EQUI7 formatting: " + str(e), exc_info=True)
                raise

            if not os.path.exists(data_equi7_fnames[unique_stack_id][0]):
                error_message = "EQUI7 grid has not been generated, output is absent "
                logging.error(error_message)
                raise

            logging.info("...done.\n")

            if stack_idx == 0:
                ### prepare the forest non forest mask
                # it is loaded if equi7, or converted to equi7 if TANDEM-X
                # it is a list containing all the loaded FNF-FTILES
                fnf_format = check_fnf_folder_format(
                    input_params_obj.stack_based_processing.forest_mask_catalogue_folder
                )
                if fnf_format == "TANDEM-X":
                    logging.info("Initial Forest mask is in TANDEM-X format, converting to equi7...")

                    # conversion step 1: get the tiff names of current zone
                    equi7_fnf_mask_fnames = fnf_tandemx_load_filter_equi7format(
                        input_params_obj.stack_based_processing.forest_mask_catalogue_folder,
                        e7g,
                        conf_params_obj.estimate_fh.product_resolution,
                        temp_output_folder,
                        gdal_path,
                        geographic_boundaries,
                        time_tag_mjd_initial,
                    )

                elif fnf_format == "EQUI7":
                    logging.info("Initial Forest mask reading and formatting...")

                    # re format the input mask in equi7 with the output resolution:
                    equi7_fnf_mask_fnames = fnf_equi7_load_filter_equi7format(
                        input_params_obj.stack_based_processing.forest_mask_catalogue_folder,
                        e7g,
                        conf_params_obj.estimate_fh.product_resolution,
                        temp_output_folder,
                        gdal_path,
                    )

            # check the equi7_fnf_mask_fnames
            check_equi7_mask_coverage(data_equi7_fnames[unique_stack_id], equi7_fnf_mask_fnames)

            # logical AND of each equi7 temp mask with each equi7 fnf mask
            mask_equi7_fnames[unique_stack_id] = fnf_and_validity_masks_merging(
                equi7_fnf_mask_fnames, validity_mask_equi7_fnames, temp_output_folder, unique_stack_id,
            )

            logging.info("...done.\n")

        # write the input file with the sections needed by the Core Processing FH APP:
        out_input_file_xml = os.path.join(
            os.path.dirname(input_params_obj.output_specification.output_folder), "Input_File_CoreProcessingFH.xml"
        )
        input_params_obj.core_processing_fh = fill_core_processing_fh_obj(
            input_params_obj, data_equi7_fnames, mask_equi7_fnames,
        )
        write_input_file(input_params_obj, out_input_file_xml)

        ######################## STACK BASED STEPS END. ###########################
        return out_input_file_xml


class CoreProcessingFH(Task):
    """CoreProcessingFH APP 
    
    CoreProcessingFH APP is the second of the two sub-APPs called by AboveGroundBiomass processor.

    It performs merging ascending with descending processed stacks and
    mosaiking together differtent processed tiles of the output grid

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file
        
    Methods
    -------
    run( input_file_path )
        run the CoreProcessingFH APP
    name : str
        name of the APP

    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query : it's the first APP to be called in the manual sequence
    StackBasedProcessingFH : it’s the APP which prepares the input for this APP

    Examples
    --------
    Manual FH chain execution
        
    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.fh.main_FH import StackBasedProcessingFH
    >>> dq_obj = dataset_query()
    >>> input_file_up1 = dq_obj.run( input_file )
    >>> sbp_obj = StackBasedProcessingFH( config_file )
    >>> input_file_up2 = sbp_obj.run( input_file_up1 )
    >>> fhp_obj = CoreProcessingFH( config_file )
    >>> fhp_obj.run( input_file_up2 )

    - input_file: path of the BioPAL input file
    - input_file_up1: same of input_file with also the "stack_based_processing" section   
    - input_file_up2: same of input_file_up1 with also the "core_processing_fh" section
    - config_file: path of the BioPAL configuration file
    """

    def __init__(self, configuration_file):

        super().__init__(configuration_file)

    def _run(self, input_file):

        # FH: Reading chains configuration files
        logging.info("FH: Reading chains configuration files")
        check_if_path_exists(self.configuration_file, "FILE")
        conf_params_obj = parse_configuration_file(self.configuration_file)
        input_params_obj = parse_input_file(input_file)

        # managing folders
        products_folder = os.path.join(input_params_obj.output_specification.output_folder, "Products")
        temp_output_folder = os.path.join(products_folder, "temp")

        ######################## NOT STACK BASED STEPS ############################

        stacks_to_merge_dict = collect_stacks_to_be_merged(input_params_obj.stack_based_processing.stack_composition)

        try:

            logging.info("FH: merging ascending with descending stacks....\n")

            merged_data_fnames, merging_folder = heigths_masking_and_merging(
                input_params_obj.core_processing_fh.data_equi7_fnames,
                input_params_obj.core_processing_fh.mask_equi7_fnames,
                stacks_to_merge_dict,
            )

            logging.info("...done.\n")

        except Exception as e:
            logging.error(e, exc_info=True)
            raise

        try:

            logging.info("FH: mosaiking equi7 tiles....\n")

            output_folder = os.path.join(products_folder, "global_FH")

            mosaiking(merging_folder, output_folder)

            logging.info("...done.\n")

        except Exception as e:
            logging.error(e, exc_info=True)
            raise

        if conf_params_obj.processing_flags.delete_temporary_files:
            try:
                shutil.rmtree(temp_output_folder)
            except:
                pass

        logging.info("FH: Forest Height estimation ended correctly.\n")


def fill_core_processing_fh_obj(input_params_obj, data_equi7_fnames, mask_equi7_fnames):

    """
    Internal function called by the StackBasedProcessingFH APP:
        
        the StackBasedProcessingFH APP fills the structure
        to be written into the xml input file for the next APP, which is the 
        core processing for the FH.
        The returned object "core_processing_FH_obj" contains the paths of 
        computed data and masks in equi7 format
        
        Usage of the returned object:
            The returned object can be added to the input_params_obj and it 
            can be written to disk if needed, as (core_processing_fh_obj is 
                                                  overwritten by this command, if already present):
                - input_params_obj.core_processing_fh_obj = core_processing_fh_obj
                - write_input_file(input_params_obj, input_file_xml)
    """

    core_processing_fh_obj = core_processing_fh(data_equi7_fnames, mask_equi7_fnames,)

    return core_processing_fh_obj



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
MetaData module
-------------------------------
"""
from abc import ABCMeta
import enum
import warnings
import collections
import copy
import numpy as np
from ..timing.precisedatetime import PreciseDateTime
from .. import constants


class EByteOrder(enum.Enum):
    """
    Byte orders supported ProductFolder's raster
    """
    be = 'BIGENDIAN'
    le = 'LITTLEENDIAN'


class ECellType(enum.Enum):
    """
    Data format supported in the ProductFolder' raster
    """
    int8 = 'INT8'
    int16 = 'INT16'
    int32 = 'INT32'
    float32 = 'FLOAT32'
    float64 = 'FLOAT64'
    i8complex = 'INT8_COMPLEX'
    i16complex = 'INT16_COMPLEX'
    i32complex = 'INT32_COMPLEX'
    fcomplex = 'FLOAT_COMPLEX'
    dcomplex = 'DOUBLE_COMPLEX'
    custom = 'CUSTOM'


class EOrbitDirection(enum.Enum):
    """
    Satellite orbit direction
    """
    ascending = 'ASCENDING'
    descending = 'DESCENDING'


class ESideLooking(enum.Enum):
    """
    Satellite side looking
    """
    right_looking = 'RIGHT'
    left_looking = 'LEFT'


class EPolarization(enum.Enum):
    """
    Polarizations
    """
    hh = 'H/H'
    vv = 'V/V'
    hv = 'H/V'
    vh = 'V/H'
    none = None
    crh = 'CR/H'
    crv = 'CR/V'
    clh = 'CL/H'
    clv = 'CL/V'
    ch = 'C/H'
    cv = 'C/V'
    xh = 'x/H'
    xv = 'x/V'
    hx = 'H/x'
    vx = 'V/x'


class EPulseDirection(enum.Enum):
    """
    Chirp pulse type
    """
    up = 'UP'
    down = 'DOWN'


class EReferenceFrame(enum.Enum):
    """
    Data reference frame
    """
    geocentric = 'GEOCENTRIC'
    geodetic = 'GEODETIC'
    zerodoppler = 'ZERODOPPLER'
    none = ''


class ERotationOrder(enum.Enum):
    """
    Attitude rotation order
    - y: yaw
    - p: pitch
    - r: roll
    """
    ypr = 'ypr'
    yrp = 'yrp'
    pry = 'pry'
    pyr = 'pyr'
    ryp = 'ryp'
    rpy = 'rpy'
    none = ''


class EAttitudeType(enum.Enum):
    """
    Attitude type
    """
    nominal = 'NOMINAL'
    refined = 'REFINED'
    none = ''


class ERasterFormatType(enum.Enum):
    """
    Aresys raster format
    """
    aresys_raster = 'ARESYS_RASTER'
    aresys_geotiff = 'ARESYS_GEOTIFF'
    raster = 'ARESYS_RASTER'


class MetaDataElement(metaclass=ABCMeta):
    """
    Base class for metadata elements
    """
    TYPE = None

    def __repr__(self):
        max_len = max([len(x) for x in self.__dict__])
        str_repr = ["\nMetaDataElement: {}\n\n".format(self.TYPE)]
        str_repr += [
            "{elemName:>{length}}: {value}\n".format(elemName=k.lstrip('_.- '), length=max_len+1,
                                                     value=v) for k, v in self.__dict__.items()]
        return "".join(str_repr)

    def type(self):
        return self.TYPE

    def copy(self):
        return copy.copy(self)


class RasterInfo(MetaDataElement):
    """
    RasterInfo class
    """
    TYPE = 'RasterInfo'

    def __init__(self, lines, samples, celltype, filename=None,
                 header_offset_bytes=0, row_prefix_bytes=0,
                 byteorder='LITTLEENDIAN', invalid_value=None, format_type=None):
        self._file_name = filename

        if isinstance(lines, int):
            self._lines = lines
        else:
            raise ValueError('The lines parameter have to be an integer')

        if isinstance(samples, int):
            self._samples = samples
        else:
            raise ValueError('The samples parameter have to be an integer')

        if isinstance(header_offset_bytes, int):
            self._header_offset_bytes = header_offset_bytes
        else:
            raise ValueError('The header offset parameter have to be an integer')

        if isinstance(row_prefix_bytes, int):
            self._row_prefix_bytes = row_prefix_bytes
        else:
            raise ValueError('The row prefix parameter have to be an integer')

        self._byte_order = EByteOrder(byteorder)
        self._cell_type = ECellType(celltype)

        self._lines_start = 0.
        self._lines_start_unit = ''
        self._lines_step = 0.
        self._lines_step_unit = ''

        self._samples_start = 0.
        self._samples_start_unit = ''
        self._samples_step = 0.
        self._samples_step_unit = ''

        self._invalid_value = invalid_value

        if format_type is not None:
            format_type = ERasterFormatType(format_type)
        self._format_type = format_type

    @property
    def file_name(self):
        return self._file_name

    @property
    def lines(self):
        return self._lines

    @property
    def samples(self):
        return self._samples

    @property
    def header_offset_bytes(self):
        return self._header_offset_bytes

    @property
    def row_prefix_bytes(self):
        return self._row_prefix_bytes

    @property
    def lines_start(self):
        return self._lines_start

    @property
    def lines_start_unit(self):
        return self._lines_start_unit

    @property
    def lines_step(self):
        return self._lines_step

    @property
    def lines_step_unit(self):
        return self._lines_step_unit

    @property
    def samples_start(self):
        return self._samples_start

    @property
    def samples_start_unit(self):
        return self._samples_start_unit

    @property
    def samples_step(self):
        return self._samples_step

    @property
    def samples_step_unit(self):
        return self._samples_step_unit

    @property
    def byte_order(self):
        return self._byte_order

    @property
    def cell_type(self):
        return self._cell_type

    @property
    def invalid_value(self):
        return self._invalid_value

    @property
    def format_type(self):
        return self._format_type

    def set_lines_axis(self, lines_start,
                       lines_start_unit,
                       lines_step,
                       lines_step_unit):
        """
        setter of the RasterInfo lines axis
        :param lines_start:
        :param lines_start_unit:
        :param lines_step:
        :param lines_step_unit:
        """
        self._lines_start = lines_start
        self._lines_start_unit = lines_start_unit
        self._lines_step = lines_step
        self._lines_step_unit = lines_step_unit

    def set_samples_axis(self, samples_start,
                         samples_start_unit,
                         samples_step,
                         samples_step_unit):
        self._samples_start = samples_start
        self._samples_start_unit = samples_start_unit
        self._samples_step = samples_step
        self._samples_step_unit = samples_step_unit


class DataSetInfo(MetaDataElement):
    """
    DataSetInfo class
    """
    TYPE = 'DataSetInfo'

    def __init__(self, acquisition_mode_i=None, fc_hz_i=None):
        self.sensor_name = None
        self.description = None
        self._sense_date = None
        self.acquisition_mode = acquisition_mode_i
        self.image_type = None
        self.projection = None
        self.acquisition_station = None
        self.processing_center = None
        self._processing_date = None
        self.processing_software = None
        self.fc_hz = fc_hz_i
        self._side_looking = None
        self.external_calibration_factor = None
        self.data_take_id = None

    @property
    def sense_date(self):
        return self._sense_date

    @sense_date.setter
    def sense_date(self, sense_date_i):
        if isinstance(sense_date_i, PreciseDateTime) or sense_date_i is None:
            self._sense_date = sense_date_i
        else:
            raise ValueError('Sense date have to be a PreciseDateTime')

    @property
    def processing_date(self):
        return self._processing_date

    @processing_date.setter
    def processing_date(self, processing_date_i):
        if isinstance(processing_date_i, PreciseDateTime) or processing_date_i is None:
            self._processing_date = processing_date_i
        else:
            raise ValueError('Processing date have to be a PreciseDateTime')

    @property
    def fc_hz(self):
        return self._fc_hz

    @fc_hz.setter
    def fc_hz(self, fc_hz_i):
        self._fc_hz = fc_hz_i

    @property
    def side_looking(self):
        return self._side_looking

    @side_looking.setter
    def side_looking(self, side_looking_i):
        self._side_looking = ESideLooking(side_looking_i)


class GeoPoint(MetaDataElement):
    """
    GeoPoint class
    """
    TYPE = 'GeoPoint'

    def __init__(self, lat=0, lon=0, height=0, theta_inc=0, theta_look=0):
        self.lat = lat
        self.lon = lon
        self.height = height
        self.theta_inc = theta_inc
        self.theta_look = theta_look

    def to_list(self):
        return [self.lat, self.lon, self.height, self.theta_inc, self.theta_look]


class GroundCornerPoints(MetaDataElement):
    """
    GrondCornerPoint class
    """
    TYPE = 'GroundCornerPoints'

    def __init__(self):
        self.easting_grid_size = 0
        self.northing_grid_size = 0
        self.geo_points = [GeoPoint() for _ in range(5)]

    @property
    def center_point(self):
        return self.geo_points[4]

    @center_point.setter
    def center_point(self, center: GeoPoint):
        self.geo_points[4] = center

    @property
    def ne_point(self):
        return self.geo_points[1]

    @ne_point.setter
    def ne_point(self, ne_point: GeoPoint):
        self.geo_points[1] = ne_point

    @property
    def nw_point(self):
        return self.geo_points[0]

    @nw_point.setter
    def nw_point(self, nw_point: GeoPoint):
        self.geo_points[0] = nw_point

    @property
    def se_point(self):
        return self.geo_points[3]

    @se_point.setter
    def se_point(self, se_point: GeoPoint):
        self.geo_points[3] = se_point

    @property
    def sw_point(self):
        return self.geo_points[2]

    @sw_point.setter
    def sw_point(self, sw_point: GeoPoint):
        self.geo_points[2] = sw_point


class SwathInfo(MetaDataElement):
    """
    SwathInfo class
    """
    TYPE = 'SwathInfo'

    def __init__(self, swath_i=None, polarization_i=None, acquisition_prf_i=0):
        self.swath = swath_i
        self.polarization = polarization_i
        self.acquisition_prf = acquisition_prf_i
        self.acquisition_prf_unit = 'Hz'
        self.swath_acquisition_order = 0
        self.rank = 0
        self.range_delay_bias = 0
        self.range_delay_bias_unit = 's'
        self.acquisition_start_time = None
        self.acquisition_start_time_unit = 'Utc'
        self.azimuth_steering_rate_reference_time = 0
        self.az_steering_rate_ref_time_unit = 's'
        self.echoes_per_burst = 0
        self.azimuth_steering_rate_pol = (0, 0, 0)
        self.rx_gain = None
        self.channel_delay = None

    @property
    def polarization(self):
        return self._polarization

    @polarization.setter
    def polarization(self, polarization_i):
        self._polarization = EPolarization(polarization_i)

    @property
    def acquisition_prf(self):
        return self._acquisition_prf

    @acquisition_prf.setter
    def acquisition_prf(self, acquisition_prf_i):
        self._acquisition_prf = acquisition_prf_i

    @property
    def acquisition_start_time(self):
        return self._acquisition_start_time

    @acquisition_start_time.setter
    def acquisition_start_time(self, acquisition_start_time_i):
        if isinstance(acquisition_start_time_i, PreciseDateTime):
            self._acquisition_start_time = acquisition_start_time_i
        elif acquisition_start_time_i is None:
            self._acquisition_start_time = None
        else:
            raise TypeError('Acquisition start time has to be a PreciseDateTime')

    @property
    def azimuth_steering_rate_pol(self):
        return self._azimuth_steering_rate_pol

    @azimuth_steering_rate_pol.setter
    def azimuth_steering_rate_pol(self, i_azimuth_steering_pol):
        if len(i_azimuth_steering_pol) == 3 and isinstance(i_azimuth_steering_pol, tuple):
            self._azimuth_steering_rate_pol = i_azimuth_steering_pol
        else:
            raise TypeError('The azimuth steering rate pol has to be a tuple of 3 elements')


class SamplingConstants(MetaDataElement):
    """
    SamplingConstants class
    """
    TYPE = 'SamplingConstants'

    def __init__(self, frg_hz_i=None, brg_hz_i=None, faz_hz_i=None, baz_hz_i=None):
        self.frg_hz = frg_hz_i
        self.frg_hz_unit = 'Hz'
        self.brg_hz = brg_hz_i
        self.brg_hz_unit = 'Hz'
        self.faz_hz = faz_hz_i
        self.faz_hz_unit = 'Hz'
        self.baz_hz = baz_hz_i
        self.baz_hz_unit = 'Hz'


class AcquisitionTimeLine(MetaDataElement):
    """
    AcquisitionTimeLine class
    """
    TYPE = 'AcquisitionTimeLine'

    def __init__(self, missing_lines_number_i=0, missing_lines_azimuth_times_i=None,
                 swst_changes_number_i=0, swst_changes_azimuth_times_i=None,
                 swst_changes_values_i=None, noise_packets_number_i=0,
                 noise_packets_azimuth_times_i=None, internal_calibration_number_i=0,
                 internal_calibration_azimuth_times_i=None, swl_changes_number_i=0,
                 swl_changes_azimuth_times_i=None, swl_changes_values_i=None):

        integer_arguments = (missing_lines_number_i,
                             swst_changes_number_i,
                             noise_packets_number_i,
                             internal_calibration_number_i,
                             swl_changes_number_i)

        integer_tags = ('missing_lines_number_i',
                        'swst_changes_number_i',
                        'noise_packets_number_i',
                        'internal_calibration_number_i',
                        'swl_changes_number_i')

        size_to_vec = dict()
        size_to_vec['missing_lines_number_i'] = (missing_lines_number_i, [missing_lines_azimuth_times_i])
        size_to_vec['swst_changes_number_i'] = (
            swst_changes_number_i, [swst_changes_azimuth_times_i, swst_changes_values_i])
        size_to_vec['noise_packets_number_i'] = (noise_packets_number_i, [noise_packets_azimuth_times_i])
        size_to_vec['internal_calibration_number_i'] = (
            internal_calibration_number_i, [internal_calibration_azimuth_times_i])
        size_to_vec['swl_changes_number_i'] = (
            swl_changes_number_i, [swl_changes_azimuth_times_i, swl_changes_values_i])

        for arg, tag in zip(integer_arguments, integer_tags):
            if not isinstance(arg, int):
                if isinstance(arg, float):
                    if abs(int(arg) - arg) == 0.0:
                        arg = int(arg)
                    else:
                        raise ValueError("{} should be an integer not a float".format(tag))
                else:
                    try:
                        arg = int(arg)
                    except TypeError as exc:
                        raise ValueError("{} wrong type: {} != int".format(tag, type(arg))) from exc
            if arg < 0:
                raise ValueError("{} should be non-negative".format(tag))

        for tag, tup in size_to_vec.items():
            for vec in tup[1]:
                if vec is not None:
                    if len(vec) != tup[0]:
                        raise ValueError("Incorrect size of vectors ({}) {} != {}".format(tag, len(vec), tup[0]))

        self._missing_lines_number = missing_lines_number_i
        self._missing_lines_azimuth_times = missing_lines_azimuth_times_i

        self._duplicated_lines_number = 0
        self._duplicated_lines_azimuth_times = None

        self._swst_changes_number = swst_changes_number_i
        self._swst_changes_azimuth_times = swst_changes_azimuth_times_i
        self._swst_changes_values = swst_changes_values_i
        self._noise_packets_number = noise_packets_number_i
        self._noise_packets_azimuth_times = noise_packets_azimuth_times_i
        self._internal_calibration_number = internal_calibration_number_i
        self._internal_calibration_azimuth_times = internal_calibration_azimuth_times_i
        self._swl_changes_number = swl_changes_number_i
        self._swl_changes_azimuth_times = swl_changes_azimuth_times_i
        self._swl_changes_values = swl_changes_values_i

        # Units:
        self.missing_lines_azimuth_times_unit = 's'
        self.duplicated_lines_azimuth_times_unit = 's'
        self.swst_changes_azimuth_times_unit = 's'
        self.swst_changes_values_unit = 's'
        self.noise_packets_azimuth_times_unit = 's'
        self.internal_calibration_azimuth_times_unit = 's'
        self.swl_changes_azimuth_times_unit = 's'
        self.swl_changes_values_unit = 's'

    @property
    def duplicated_lines(self):
        return self._duplicated_lines_number, self._duplicated_lines_azimuth_times

    @duplicated_lines.setter
    def duplicated_lines(self, duplicated_lines):
        if len(duplicated_lines[1]) != duplicated_lines[0]:
            raise ValueError(
                "Duplicated lines inconsistent tuple: {} != {}".format(duplicated_lines[0], len(duplicated_lines[1])))

        self._duplicated_lines_number = duplicated_lines[0]
        self._duplicated_lines_azimuth_times = duplicated_lines[1]

    @property
    def internal_calibration(self):
        return self._internal_calibration_number, self._internal_calibration_azimuth_times

    @internal_calibration.setter
    def internal_calibration(self, internal_calibration):
        self._internal_calibration_number = len(internal_calibration)
        self._internal_calibration_azimuth_times = internal_calibration

    @property
    def missing_lines(self):
        return self._missing_lines_number, self._missing_lines_azimuth_times

    @missing_lines.setter
    def missing_lines(self, azimuth_times):
        self._missing_lines_number = len(azimuth_times)
        self._missing_lines_azimuth_times = azimuth_times

    @property
    def noise_packet(self):
        return self._noise_packets_number, self._noise_packets_azimuth_times

    @noise_packet.setter
    def noise_packet(self, azimuth_times):
        self._noise_packets_number = len(azimuth_times)
        self._noise_packets_azimuth_times = azimuth_times

    @property
    def swl_changes(self):
        return self._swl_changes_number, self._swl_changes_azimuth_times, self._swl_changes_values

    @swl_changes.setter
    def swl_changes(self, swl_changes):
        if swl_changes[0] != len(swl_changes[1]):
            raise ValueError("Inconsistent swl changes sizes")
        self._swl_changes_number = swl_changes[0]
        self._swl_changes_azimuth_times = swl_changes[1]
        self._swl_changes_values = swl_changes[2]

    @property
    def swst_changes(self):
        return self._swst_changes_number, self._swst_changes_azimuth_times, self._swst_changes_values

    @swst_changes.setter
    def swst_changes(self, swst_changes):
        if swst_changes[0] != len(swst_changes[1]):
            raise ValueError("Inconsistent swl changes sizes")
        self._swst_changes_number = swst_changes[0]
        self._swst_changes_azimuth_times = swst_changes[1]
        self._swst_changes_values = swst_changes[2]


class AttitudeInfo(MetaDataElement):
    """
    AttitudeInfo class
    """

    TYPE = 'AttitudeInfo'

    _default_attitude_type = EAttitudeType('NOMINAL')

    def __init__(self, yaw=None, pitch=None, roll=None, t0=None, delta_t=0, ref_frame=None, rot_order=None):
        if ref_frame is None:
            ref_frame = ''
        if rot_order is None:
            rot_order = ''
        if t0 is None:
            t0 = PreciseDateTime()

        self._reset_angles(yaw, pitch, roll)

        self._reference_frame = None
        self.reference_frame = ref_frame
        self._rotation_order = None
        self.rotation_order = rot_order

        self._attitude_type = self._default_attitude_type

        self._t_ref_Utc = None
        self.reference_time = t0

        self._dtYPR_s = None
        self.time_step = delta_t

    def _reset_angles(self, yaw, pitch, roll):
        if yaw is None and pitch is None and roll is None:
            return
        yaw_vector = np.array(yaw)
        pitch_vector = np.array(pitch)
        roll_vector = np.array(roll)

        yaw_size = yaw_vector.size
        for vec, tag in zip((yaw_vector, pitch_vector, roll_vector), ('yaw', 'pitch', 'roll')):
            wrong_dimension_string = 'Provided {} vector shall be a 1xN or Nx1 array'.format(tag)
            if vec.ndim == 1:
                if vec.size != yaw_size:
                    raise ValueError(wrong_dimension_string)
            elif vec.ndim == 2:
                if vec.shape[0] != 1 and vec.shape[1] != 1:
                    raise ValueError(wrong_dimension_string)
            else:
                raise ValueError(wrong_dimension_string)

            if yaw_size != vec.size:
                raise ValueError('{} and yaw vectors must have the same number of elements'.format(tag))
        self._yaw_deg = yaw_vector
        self._pitch_deg = pitch_vector
        self._roll_deg = roll_vector
        self._nYPR_n = yaw_size

    @property
    def reference_frame(self) -> EReferenceFrame:
        return self._reference_frame

    @reference_frame.setter
    def reference_frame(self, reference_frame: str):
        self._reference_frame = EReferenceFrame(reference_frame.upper())

    @property
    def rotation_order(self) -> ERotationOrder:
        return self._rotation_order

    @rotation_order.setter
    def rotation_order(self, rotation_order: str):
        self._rotation_order = ERotationOrder(rotation_order.lower())

    @property
    def attitude_records_number(self):
        return self._nYPR_n

    @property
    def attitude_type(self) -> EAttitudeType:
        return self._attitude_type

    @attitude_type.setter
    def attitude_type(self, attitude_type: str):
        self._attitude_type = EAttitudeType(attitude_type.upper())

    @property
    def pitch_vector(self):
        return self._pitch_deg

    @property
    def reference_time(self):
        return self._t_ref_Utc

    @reference_time.setter
    def reference_time(self, reference_time):
        if not isinstance(reference_time, (int, float, PreciseDateTime)):
            raise ValueError('Input start time must be a scalar PreciseDateTime')
        self._t_ref_Utc = reference_time

    @property
    def roll_vector(self):
        return self._roll_deg

    @property
    def time_step(self):
        return self._dtYPR_s

    @time_step.setter
    def time_step(self, delta_t):
        self._dtYPR_s = delta_t

    @property
    def yaw_vector(self):
        return self._yaw_deg

    def set_attitude_angles_vectors(self, yaw, pitch, roll):
        self._reset_angles(yaw, pitch, roll)


class Burst(MetaDataElement):
    """
    Single Burst class
    """
    TYPE = 'Burst'

    def __init__(self, range_start_time_i, azimuth_start_time_i, lines_i,
                 burst_center_azimuth_shift_i=None):
        self._range_start_time = range_start_time_i
        self._azimuth_start_time = azimuth_start_time_i
        self._burst_center_azimuth_shift = burst_center_azimuth_shift_i
        self._lines = lines_i

    @property
    def range_start_time(self):
        return self._range_start_time

    @property
    def azimuth_start_time(self):
        return self._azimuth_start_time

    @property
    def burst_center_azimuth_shift(self):
        return self._burst_center_azimuth_shift

    @property
    def lines(self):
        return self._lines


class BurstInfo(MetaDataElement):
    """
    BurstInfo class
    """

    TYPE = 'BurstInfo'

    def __init__(self, burst_repetition_frequency=0):
        self._bursts = list()
        self._lines_per_burst_present = False
        self._lines_per_burst = 0
        self.burst_repetition_frequency = burst_repetition_frequency

    def is_lines_per_burst_present(self):
        return self._lines_per_burst_present

    @property
    def lines_per_burst(self):
        return self._lines_per_burst

    def add_burst(self, range_start_time_i, azimuth_start_time_i, lines_i,
                  burst_center_azimuth_shift_i=None):

        if len(self._bursts) == 0:
            self._lines_per_burst_present = True
            self._lines_per_burst = lines_i
        else:
            if self._lines_per_burst_present:
                if self._lines_per_burst != lines_i:
                    self._lines_per_burst_present = False

        if burst_center_azimuth_shift_i is None:
            burst_ = Burst(range_start_time_i, azimuth_start_time_i, lines_i)
        else:
            burst_ = Burst(range_start_time_i, azimuth_start_time_i, lines_i,
                           burst_center_azimuth_shift_i)
        self._bursts.append(burst_)

    def get_number_of_bursts(self):
        return len(self._bursts)

    def get_burst(self, burst_index):
        if 0 <= burst_index < self.get_number_of_bursts():
            return self._bursts[burst_index]

        raise RuntimeError('Burst index out of range')

    def clear_bursts(self):
        self._bursts = list()

    def _get_bursts_property(self, burst_index, attr):
        if burst_index is None:
            return np.asarray([getattr(burst, attr) for burst in self._bursts])

        if burst_index < 0 or burst_index >= self.get_number_of_bursts():
            raise ValueError("Not valid butst index")

        return getattr(self._bursts[burst_index], attr)

    def get_lines(self, burst_index=None):
        return self._get_bursts_property(burst_index, 'lines')

    def get_azimuth_start_time(self, burst_index=None):
        return self._get_bursts_property(burst_index, 'azimuth_start_time')

    def get_burst_center_azimuth_shift(self, burst_index=None):
        return self._get_bursts_property(burst_index, 'burst_center_azimuth_shift')

    def get_range_start_time(self, burst_index=None):
        return self._get_bursts_property(burst_index, 'range_start_time')

    def get_burst_roi(self, burst_index, raster_info: RasterInfo, roi_range=None):
        if roi_range is None:
            roi_range = [1, raster_info.samples]

        accumulated_lines = 0
        for i_burst in range(0, burst_index):
            accumulated_lines += self.get_lines(i_burst)

        first_line = accumulated_lines
        return [first_line, roi_range[0], self.get_lines(burst_index), roi_range[2]]


class StateVectors(MetaDataElement):
    """
    StateVectors class
    """
    TYPE = 'StateVectors'

    def __init__(self, position_vector: np.ndarray = None,
                 velocity_vector: np.ndarray = None,
                 t_ref_utc=None,
                 dt_sv_s=0):
        position_vector = np.array(position_vector)
        velocity_vector = np.array(velocity_vector)

        if position_vector.ndim > 2 or position_vector.size % 3 != 0:
            raise ValueError('Wrong array size for input position vector')
        if velocity_vector.ndim > 2 or velocity_vector.size % 3 != 0:
            raise ValueError('Wrong array size for input velocity vector')

        self._position_vector = position_vector
        self._velocity_vector = velocity_vector
        self._t_ref_utc = t_ref_utc
        self._dt_sv_s = dt_sv_s
        self._orbit_number = -1
        self._track_number = -1
        self._anx_time = None
        self._anx_position = None

    @property
    def anx_position(self):
        return self._anx_position

    def get_anx_time(self):
        return self._anx_time

    @property
    def orbit_direction(self):
        if self._velocity_vector[0][2] > 0:
            return EOrbitDirection.ascending

        return EOrbitDirection.descending

    @property
    def orbit_number(self):
        return self._orbit_number

    @property
    def position_vector(self):
        return self._position_vector

    @property
    def reference_time(self):
        return self._t_ref_utc

    @property
    def number_of_state_vectors(self) -> int:
        return int(self._position_vector.size / 3)

    @property
    def time_step(self):
        return self._dt_sv_s

    @property
    def track_number(self):
        return self._track_number

    @property
    def velocity_vector(self):
        return self._velocity_vector

    def set_axn_info(self, i_anx_time, i_anx_pos):
        if len(i_anx_pos) != 3 or not isinstance(i_anx_pos, list):
            raise ValueError('Wrong input type for ANX position. It must be a list of 3 numbers.')
        self._anx_position = i_anx_pos

        if len(i_anx_time) != 1 or not isinstance(i_anx_time, str):
            raise ValueError('Wrong input type for ANX time. It must be an Mjd type.')
        self._anx_time = i_anx_time

    @orbit_number.setter
    def orbit_number(self, i_orbit_number):
        if i_orbit_number <= 0 or not isinstance(i_orbit_number, int):
            raise ValueError('Provided orbit number must have an integer and positive value.')
        self._orbit_number = i_orbit_number

    def set_state_vectors(self, i_position, i_velocity, i_reference_time, i_time_step):
        raise NotImplementedError('')

    @track_number.setter
    def track_number(self, i_track_number):
        if not isinstance(i_track_number, int) or i_track_number <= 0:
            raise ValueError('Provided track number must have an integer and positive value.')
        self._track_number = i_track_number


class _Poly2D(MetaDataElement):
    """
    Base class for Poly2D format
    """
    TYPE = 'Poly2D'
    _POWERS_X = (0,
                 0,
                 1,
                 1,
                 0,
                 0,
                 0,
                 0,
                 0,
                 0,
                 0)

    _POWERS_Y = (0,
                 1,
                 0,
                 1,
                 2,
                 3,
                 4,
                 5,
                 6,
                 7,
                 8)

    _UNITS = ('Hz',
              'Hz/s',
              'Hz/s',
              'Hz/s2',
              'Hz/s2',
              'Hz/s3',
              'Hz/s4',
              'Hz/s5',
              'Hz/s6',
              'Hz/s7',
              'Hz/s8',)

    def __init__(self, i_ref_az=None, i_ref_rg=None,
                 i_coefficients=None):
        self.t_ref_az = i_ref_az
        self.t_ref_rg = i_ref_rg
        if len(i_coefficients) > len(self._POWERS_X):
            raise ValueError('the size of coefficients and powers must agree')
        self._coefficients = i_coefficients

    @property
    def coefficients(self):
        return self._coefficients

    @staticmethod
    def get_powers_x():
        return _Poly2D._POWERS_X

    @staticmethod
    def get_powers_y():
        return _Poly2D._POWERS_Y

    @staticmethod
    def get_units():
        return _Poly2D._UNITS


class _Poly2DVector(MetaDataElement):
    """
    Base class for list of Poly2D
    """
    TYPE = 'Poly2DVector'
    _SINGLE_POLY_TYPE = _Poly2D

    def __init__(self, i_poly2d: list = None):
        self._poly_list = i_poly2d if i_poly2d is not None else list()
        assert isinstance(self._poly_list, list), "The input should be a list"
        self._current_poly_index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._current_poly_index >= self.get_number_of_poly():
            self._current_poly_index = 0
            raise StopIteration
        self._current_poly_index += 1
        return self.get_poly(self._current_poly_index - 1)

    def add_poly(self, i_poly2d):
        self._poly_list.append(i_poly2d)

    def get_number_of_poly(self):
        return len(self._poly_list)

    def get_poly(self, index):
        if 0 <= index < self.get_number_of_poly():
            return self._poly_list[index]

        raise IndexError('Polynomial not available for index: {}'.format(index))

    @classmethod
    def get_single_poly_type(cls):
        return cls._SINGLE_POLY_TYPE


class DopplerCentroid(_Poly2D):
    """
    DopplerCentroid class
    """
    TYPE = 'DopplerCentroid'


class DopplerCentroidVector(_Poly2DVector):
    """
    List of DopplerCentroid poly
    """
    TYPE = 'DopplerCentroidVector'
    _SINGLE_POLY_TYPE = DopplerCentroid


class DopplerRate(_Poly2D):
    """
    DopplerRate class
    """
    TYPE = 'DopplerRate'


class DopplerRateVector(_Poly2DVector):
    """
    list of DopplerRate poly
    """
    TYPE = 'DopplerRateVector'
    _SINGLE_POLY_TYPE = DopplerRate


class TopsAzimuthModulationRate(_Poly2D):
    """
    TopsAzimuthModulationRate class
    """
    TYPE = 'TopsAzimuthModulationRate'


class TopsAzimuthModulationRateVector(_Poly2DVector):
    """
    List of TopsAzimuthModulationRate poly
    """
    TYPE = 'TopsAzimuthModulationRateVector'
    _SINGLE_POLY_TYPE = TopsAzimuthModulationRate


class SlantToGround(_Poly2D):
    """
    SlantToGround class
    """
    TYPE = 'SlantToGround'


class SlantToGroundVector(_Poly2DVector):
    """
    List of SlantToGround poly
    """
    TYPE = 'SlantToGroundVector'
    _SINGLE_POLY_TYPE = SlantToGround


class GroundToSlant(_Poly2D):
    """
    GroundToSlant class
    """
    TYPE = 'GroundToSlant'


class GroundToSlantVector(_Poly2DVector):
    """
    List of GroundToSlant poly
    """
    TYPE = 'GroundToSlantVector'
    _SINGLE_POLY_TYPE = GroundToSlant


class SlantToIncidence(_Poly2D):
    """
    SlantToIncidence class
    """
    TYPE = 'SlantToIncidence'


class SlantToIncidenceVector(_Poly2DVector):
    """
    List of SlantToIncidence poly
    """
    TYPE = 'SlantToIncidenceVector'
    _SINGLE_POLY_TYPE = SlantToIncidence


class SlantToElevation(_Poly2D):
    """
    SlantToElevation class
    """
    TYPE = 'SlantToElevation'


class SlantToElevationVector(_Poly2DVector):
    """
    List of SlantToElevation poly
    """
    TYPE = 'SlantToElevationVector'
    _SINGLE_POLY_TYPE = SlantToElevation


class AntennaInfo(MetaDataElement):
    """
    AntennaInfo class
    """
    TYPE = 'AntennaInfo'

    def __init__(self, i_sensor_name=None, i_polarization=None, i_acquisition_mode=None,
                 i_acquisition_beam=None, i_lines_per_pattern=0):
        self.sensor_name = i_sensor_name
        self.polarization = i_polarization
        self.acquisition_mode = i_acquisition_mode
        self.acquisition_beam = i_acquisition_beam
        self.lines_per_pattern = i_lines_per_pattern

    @property
    def polarization(self):
        return self._polarization

    @polarization.setter
    def polarization(self, i_polarization):
        self._polarization = EPolarization(i_polarization)


class DataStatistics(MetaDataElement):
    """
    DataStatistics class
    """
    TYPE = 'DataStatistics'

    def __init__(self, i_num_samples=0, i_max_i=0,
                 i_max_q=0, i_min_i=0, i_min_q=0,
                 i_sum_i=0, i_sum_q=0, i_sum_2_i=0,
                 i_sum_2_q=0, i_std_dev_i=0, i_std_dev_q=0):
        self.num_samples = i_num_samples
        self.max_i = i_max_i
        self.max_q = i_max_q
        self.min_i = i_min_i
        self.min_q = i_min_q
        self.sum_i = i_sum_i
        self.sum_q = i_sum_q
        self.sum_2_i = i_sum_2_i
        self.sum_2_q = i_sum_2_q
        self.std_dev_i = i_std_dev_i
        self.std_dev_q = i_std_dev_q
        self._statistics_list = list()

    def add_data_block_statistic(self, i_data_block_statistic):
        if isinstance(i_data_block_statistic, DataBlockStatistic):
            self._statistics_list.append(i_data_block_statistic)
        else:
            raise TypeError('The input must be of type DataBlockStatistic')

    def get_data_block_statistic(self, index):
        if 0 <= index < self.get_number_of_data_block_statistic():
            return self._statistics_list[index]

        raise IndexError('DataBlockStatistic not available for index: {}'.format(index))

    def get_number_of_data_block_statistic(self):
        return len(self._statistics_list)


class DataBlockStatistic(MetaDataElement):
    """
    DataBlockStatistic class
    """
    TYPE = 'DataBlockStatistic'

    def __init__(self, i_line_start, i_line_stop,
                 i_num_samples=0, i_max_i=0,
                 i_max_q=0, i_min_i=0, i_min_q=0,
                 i_sum_i=0, i_sum_q=0, i_sum_2_i=0,
                 i_sum_2_q=0):
        self.num_samples = i_num_samples
        self.max_i = i_max_i
        self.max_q = i_max_q
        self.min_i = i_min_i
        self.min_q = i_min_q
        self.sum_i = i_sum_i
        self.sum_q = i_sum_q
        self.sum_2_i = i_sum_2_i
        self.sum_2_q = i_sum_2_q
        self.line_start = i_line_start
        self.line_stop = i_line_stop


class CoregPoly(MetaDataElement):
    """
    CoregPoly class
    """
    TYPE = 'CoregPoly'

    def __init__(self, i_ref_az, i_ref_rg, i_coefficients_az, i_coefficients_rg):
        self._azimuth_coreg_poly = _Poly2D(i_ref_az, i_ref_rg, i_coefficients_az)
        self._range_coreg_poly = _Poly2D(i_ref_az, i_ref_rg, i_coefficients_rg)

    @property
    def ref_azimuth_time(self):
        return self._azimuth_coreg_poly.t_ref_az

    @ref_azimuth_time.setter
    def ref_azimuth_time(self, i_ref_az):
        self._azimuth_coreg_poly.t_ref_az = i_ref_az
        self._range_coreg_poly.t_ref_az = i_ref_az

    @property
    def ref_range_time(self):
        return self._azimuth_coreg_poly.t_ref_rg

    @ref_range_time.setter
    def ref_range_time(self, i_ref_rg):
        self._azimuth_coreg_poly.t_ref_rg = i_ref_rg
        self._range_coreg_poly.t_ref_rg = i_ref_rg

    @property
    def azimuth_poly(self):
        return self._azimuth_coreg_poly

    @property
    def range_poly(self):
        return self._range_coreg_poly


class CoregPolyVector(_Poly2DVector):
    """
    List of CoregPoly
    """
    TYPE = 'CoregPolyVector'
    _SINGLE_POLY_TYPE = CoregPoly


class Pulse(MetaDataElement):
    """
    Pulse class
    """
    TYPE = 'Pulse'

    def __init__(self, i_pulse_length=None, i_bandwidth=None, i_pulse_sampling_rate=None,
                 i_pulse_energy=None, i_pulse_start_frequency=None, i_pulse_start_phase=None,
                 i_pulse_direction=None):
        self.pulse_length = i_pulse_length
        self.pulse_length_unit = constants.SECOND_STR
        self.bandwidth = i_bandwidth
        self.bandwidth_unit = constants.HERTZ_STR
        self.pulse_energy = i_pulse_energy
        self.pulse_energy_unit = constants.JOULE_STR
        self.pulse_sampling_rate = i_pulse_sampling_rate
        self.pulse_sampling_rate_unit = constants.HERTZ_STR
        self.pulse_start_frequency = i_pulse_start_frequency
        self.pulse_start_frequency_unit = constants.HERTZ_STR
        self.pulse_start_phase = i_pulse_start_phase
        self.pulse_start_phase_unit = constants.RAD_STR
        self.pulse_direction = i_pulse_direction

    @property
    def pulse_direction(self):
        return self._pulse_direction

    @pulse_direction.setter
    def pulse_direction(self, i_pulse_direction):
        if i_pulse_direction is not None:
            self._pulse_direction = EPulseDirection(i_pulse_direction)
        else:
            self._pulse_direction = i_pulse_direction


class MetaDataChannel:
    """
    MetaDataChannel class
    """
    _supported_elements = [
        'RasterInfo',
        'DataSetInfo',
        'SwathInfo',
        'SamplingConstants',
        'AcquisitionTimeLine',
        'BurstInfo',
        'StateVectors',
        'AttitudeInfo',
        'Pulse',
        'GroundCornerPoints',
        'BurstInfo',
        'DopplerCentroidVector',
        'DopplerRateVector',
        'TopsAzimuthModulationRateVector',
        'SlantToGroundVector',
        'GroundToSlantVector',
        'SlantToIncidenceVector',
        'SlantToElevationVector',
        'AntennaInfo',
        'DataStatistics',
        'CoregPolyVector',
    ]

    def __init__(self):
        self._elements = collections.OrderedDict()
        for element_tag in self._supported_elements:
            self._elements[element_tag] = None
            setattr(self, element_tag, property(lambda s: s.get_element(element_tag)))

    def __repr__(self):
        str_repr = ['\nMetaDataChannel\n\n']
        str_repr += [str(e) for e in self._elements.values()]
        return "".join(str_repr)

    def insert_element(self, element):
        if isinstance(element, MetaDataElement):
            if self._elements[element.type()] is None:
                self._elements[element.type()] = element
            else:
                warnings.warn("The element {} is already present in the current metadata channel".format(
                    element.type()))
        else:
            raise TypeError

    def get_element(self, element_type):
        return self._elements[element_type]

    @staticmethod
    def get_supported_metadata_elements():
        return MetaDataChannel._supported_elements.copy()


class MetaData:
    """
    Metadata class

    List of MetaDataChannels
    """

    def __init__(self):
        self._metadatachannels = list()

    def append_channel(self, channel):
        if isinstance(channel, MetaDataChannel):
            self._metadatachannels.append(channel)
        else:
            raise TypeError

    def get_metadata_channels(self, channel_index):
        return self._metadatachannels[channel_index]

    def get_number_of_channels(self):
        return len(self._metadatachannels)



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
General SAR orbit module
------------------------
"""

import os
from typing import Union
import numpy as np

from arepytools.math import axis as are_ax
from arepytools import _utils
from arepytools.timing.precisedatetime import PreciseDateTime
from arepytools.io.metadata import StateVectors
from arepytools.geometry._interpolator import GeometryInterpolator
from arepytools.geometry._geocoding import _direct_geocoding, _inverse_geocoding


class GeneralSarOrbit:
    """
    GeneralSarOrbit class
    """

    _MINIMUM_NUMBER_OF_DATA_POINTS = GeometryInterpolator.get_min_num_of_data_points()

    @classmethod
    def get_minimum_number_of_data_points(cls):
        """Return the required minimum number of orbit data points

        :return: required minimum number of orbit data points
        """
        return cls._MINIMUM_NUMBER_OF_DATA_POINTS

    @property
    def t0(self):
        """Origin of the time axis"""
        return self._time_axis.start

    @property
    def dt(self):
        """Time axis step (if applicable)"""
        if isinstance(self._time_axis, are_ax.RegularAxis):
            return self._time_axis.step
        raise RuntimeError("Time step is not available for orbits constructed with non-regular time axis")

    @property
    def n(self):
        """Number of orbit data points"""
        return self._time_axis.size

    @property
    def position_sv(self):
        """State vectors as 1D array of size 3N"""
        return self._state_vectors.reshape((self.n * 3,), order='F')

    @property
    def time_axis_array(self):
        """Time axis as array of time points"""
        return self._time_axis.get_array()

    @property
    def interpolator(self):
        """Geometry interpolator object"""
        return self._interpolator

    def __init__(self, time_axis: Union[np.ndarray, are_ax.Axis], state_vectors: np.ndarray):
        """Initialize the general sar orbit object using the provided input parameters

        :param time_axis: time axis of length N, as Axis or numpy array of PreciseDateTime objects.
        :param state_vectors: state vectors as 1D numpy array of size 3N
        """

        if isinstance(time_axis, np.ndarray):
            if time_axis.dtype == PreciseDateTime:
                time_axis_start = time_axis[0]
                relative_time_axis = (time_axis - time_axis_start).astype(float)
                self._time_axis = are_ax.Axis(relative_time_axis, time_axis_start)
            else:
                raise ValueError("Axis should be a vector of PreciseDateTime objects")
        else:
            self._time_axis = time_axis

        _check_init_input(self._time_axis, state_vectors)

        # state_vector are stored as (3, N) numpy array
        self._state_vectors = np.vstack((state_vectors[::3], state_vectors[1::3], state_vectors[2::3]))
        self._interpolator = GeometryInterpolator(self._time_axis, self._state_vectors)

    def get_position(self, time_points, interval_indexes=None):
        """Return the sensor positions at the specified time points

        :param time_points: 1D numpy array of length N of absolute time points
        :param interval_indexes: intervals of the time axis where the given time points are expected [optional]

        :return: 3xN numpy array of sensor positions
        """
        return self.interpolator.eval(time_points, interval_indexes)

    def get_velocity(self, time_points, interval_indexes=None):
        """Return the sensor velocities at the specified time points, evaluated using the first derivative of the
        position data

        :param time_points: 1D numpy array of length N of absolute time points
        :param interval_indexes: intervals of the time axis where the given time points are expected [optional]

        :return: 3xN numpy array of sensor velocity vectors
        """
        return self.interpolator.eval_first_derivative(time_points, interval_indexes)

    def get_acceleration(self, time_points, interval_indexes=None):
        """Return the sensor accelerations at the specified time points, evaluated using the second derivative of the
        position data

        :param time_points: 1D numpy array of length N of absolute time points
        :param interval_indexes: intervals of the time axis where the given time points are expected [optional]

        :return: 3xN numpy array of sensor acceleration vectors
        """
        return self.interpolator.eval_second_derivative(time_points, interval_indexes)

    def sat2earth(self, time_point, range_times, look_direction, geodetic_altitude=0.,
                  doppler_centroid=None, carrier_wavelength=None, orbit_tx=None, bistatic_delay=True):
        """Compute monostatic or bistatic sensor-to-earth projection

        :param time_point: absolute time point
        :param range_times: a vector (N, 1) of range times or a single range time, as numpy array or scalar
        :param look_direction: either 'LEFT' or 'RIGHT'
        :param geodetic_altitude: geodetic altitude over WGS84 [optional]
        :param doppler_centroid: doppler centroid frequency as scalar or numpy array of size (N X 1) [optional]
        :param carrier_wavelength: sensor carrier wavelength [optional]
        :param orbit_tx: orbit of the TX sensor (for bistatic case) [optional]
        :param bistatic_delay: set it to false to evaluate TX sensor position at the RX time (for bistatic case)
                               [optional]

        :return: 3xN numpy array of earth positions in xyz coordinates
        """
        range_times_checked, doppler_centroid_checked = _check_sat2earth_input(time_point, range_times,
                                                                               doppler_centroid, carrier_wavelength)

        doppler_centroid_array = np.zeros(
            range_times_checked.shape) if doppler_centroid_checked is None else doppler_centroid_checked
        carrier_wavelength = 1. if carrier_wavelength is None else carrier_wavelength

        position_rx = self.get_position(time_point).squeeze()
        velocity_rx = self.get_velocity(time_point).squeeze()
        if orbit_tx is None:
            return _direct_geocoding.direct_geocoding_monostatic(position_rx, velocity_rx, range_times_checked,
                                                                 look_direction,
                                                                 geodetic_altitude, doppler_centroid_array,
                                                                 carrier_wavelength)
        else:
            if bistatic_delay:
                tx_time_points = [time_point - range_time for range_time in range_times_checked]
            else:
                tx_time_points = [time_point for _ in range_times_checked]

            positions_tx = orbit_tx.get_position(tx_time_points)
            velocities_tx = orbit_tx.get_velocity(tx_time_points)

            return _direct_geocoding.direct_geocoding_bistatic(position_rx, velocity_rx, positions_tx, velocities_tx,
                                                               range_times_checked, look_direction, geodetic_altitude,
                                                               doppler_centroid_array, carrier_wavelength)

    def earth2sat(self, earth_point, doppler_centroid=None, carrier_wavelength=None, orbit_tx=None):
        """Compute monostatic or bistatic earth-to-sat projection

        :param earth_point: xyz coordinates of the point on earth as 3D numpy array
        :param doppler_centroid: doppler centroid frequency [optional]
        :param carrier_wavelength: sensor carrier wavelength [optional]
        :param orbit_tx: orbit of the TX sensor (for bistatic case) [optional]

        :return: sat projections as two lists of azimuth_times and range_times
        """
        if doppler_centroid is None:
            doppler_centroid = 0.0
        if carrier_wavelength is None:
            carrier_wavelength = 1.0

        _check_earth2sat_input(earth_point)

        if orbit_tx is None:
            return _inverse_geocoding.inverse_geocoding_monostatic(self, earth_point, doppler_centroid,
                                                                   carrier_wavelength)
        else:
            return _inverse_geocoding.inverse_geocoding_bistatic(self, earth_point, doppler_centroid,
                                                                 carrier_wavelength,
                                                                 orbit_tx)

    def evaluate_doppler_equation(self, earth_point: np.ndarray, doppler_centroid, carrier_wavelength):
        """Evaluate the doppler equation over the orbit time axis

        .. math::

            \\frac{2}{\\lambda} \\frac{(P_{sat}(t) - P_0) \\cdot V_{sat}(t)}{||P_{sat}(t) - P_0||} - f_{DC}

        :param earth_point: xyz coordinates of the point on earth as 3D numpy array
        :param doppler_centroid: doppler centroid frequency
        :param carrier_wavelength: sensor carrier wavelength

        :return: resulting doppler equation values in a numpy array of size N
        """
        return doppler_equation(earth_point, self._state_vectors, self.get_velocity(self._time_axis.get_array()),
                                doppler_centroid, carrier_wavelength)

    def __repr__(self):
        axis_str = str(self._time_axis)
        state_vec_str = str(self._state_vectors)

        axis_portion = "Orbit defined on azimuth axis: " + os.linesep + axis_str + os.linesep
        state_vectors_portion = "State vectors: " + os.linesep + state_vec_str + os.linesep
        return axis_portion + state_vectors_portion

    def get_interpolated_time_axis(self, interpolation_positions):
        """Return the interpolated time axis on the given interpolation positions"""
        return self._time_axis.interpolate(interpolation_positions)


def create_general_sar_orbit(state_vectors: StateVectors):
    """Create general sar orbit object from state vectors metadata

    :param state_vectors: state vectors as a StateVectors metadata object

    :return: the new GeneralSarOrbit object
    """
    time_axis = are_ax.RegularAxis((0, state_vectors.time_step, state_vectors.number_of_state_vectors),
                                   state_vectors.reference_time)
    return GeneralSarOrbit(time_axis, state_vectors.position_vector.reshape((state_vectors.position_vector.size,)))


def doppler_equation(point, sensor_position, sensor_velocity, frequency_doppler_centroid, wavelength):
    """Evaluate doppler equation"""
    point2sensor = point - sensor_position
    distance = np.linalg.norm(point2sensor, axis=0)

    def col_wise_scalar_product(matrix_a, matrix_b):
        return np.einsum('ij,ij->j', matrix_a, matrix_b)  # Einstein notation -- col wise dot product.

    return np.divide(2 / wavelength * col_wise_scalar_product(point2sensor, sensor_velocity),
                     distance) - frequency_doppler_centroid


def _check_init_input(axis, state_vectors: np.ndarray):
    if not isinstance(axis, are_ax.Axis):
        raise RuntimeError("Axis should be of type {} != {}".format(are_ax.Axis, type(axis)))
    if not isinstance(axis.start, PreciseDateTime):
        raise RuntimeError("Wrong axis start: {} != {}".format(type(axis.start), PreciseDateTime))

    _utils.check_type(state_vectors, np.ndarray, "state_vectors")

    if axis.size != state_vectors.size / 3:
        raise RuntimeError("Size of state vectors not compatible with size of time axis: {} != {}".format(
            axis.size, state_vectors.size / 3))

    if axis.size < GeneralSarOrbit.get_minimum_number_of_data_points():
        raise RuntimeError(
            "Not enough state vectors provided. {} < {}".format(axis.size,
                                                                GeneralSarOrbit.get_minimum_number_of_data_points()))

    _utils.check_dtype_of_numpy_array(state_vectors, float, name='state_vectors')
    _utils.check_shape_of_numpy_array(state_vectors, (state_vectors.size,), name='state_vectors')


def _check_sat2earth_input(azimuth_time, range_times, frequency_doppler_centroid, wavelength):
    if not isinstance(azimuth_time, PreciseDateTime):
        raise RuntimeError("Azimuth should be a single absolute time")

    range_times = _utils.input_data_to_numpy_array_with_checks(range_times, dtype=float)

    if (frequency_doppler_centroid is not None and wavelength is None) or (
            wavelength is not None and frequency_doppler_centroid is None):
        raise RuntimeError("Frequency doppler centroid and wavelength should be both specified")

    if frequency_doppler_centroid is not None:
        frequency_doppler_centroid = _utils.input_data_to_numpy_array_with_checks(frequency_doppler_centroid,
                                                                                  dtype=float, ndim=1)

    if frequency_doppler_centroid is not None and (
            frequency_doppler_centroid.ndim != 1 or frequency_doppler_centroid.shape != range_times.shape):
        if frequency_doppler_centroid.size == 1:
            frequency_doppler_centroid = np.full(range_times.shape, frequency_doppler_centroid[0])
        else:
            raise RuntimeError(
                "Frequency doppler centroid vector should have the same shape of the range times vector: "
                + "{} != {}".format(frequency_doppler_centroid.shape, range_times.shape))

    return range_times, frequency_doppler_centroid


def _check_earth2sat_input(point):
    _utils.check_type(point, np.ndarray)
    _utils.check_dtype_of_numpy_array(point, float)
    _utils.check_shape_of_numpy_array(point, shape=(3,))


def _check_numpy_array_of_mjd(vec):
    _utils.check_type(vec, np.ndarray)
    _utils.check_ndim_of_numpy_array(vec, 1)
    _utils.check_dtype_of_numpy_array(vec, PreciseDateTime)
